/*
 * Copyright (C) 2006-2011 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file osx_write.c paraslash's output plugin for MacOs */

/*
 * based on mosx-mpg123, by Guillaume Outters and Steven A. Kortze
 * <skortze@sourceforge.net>
 */

#include <regex.h>
#include <sys/types.h>
#include <dirent.h>
#include <stdbool.h>

#include "para.h"
#include "fd.h"
#include "string.h"
#include "list.h"
#include "sched.h"
#include "ggo.h"
#include "buffer_tree.h"
#include "write.h"
#include "write_common.h"
#include "osx_write.cmdline.h"
#include "error.h"

#include <CoreServices/CoreServices.h>
#include <AudioUnit/AudioUnit.h>
#include <AudioToolbox/AudioToolbox.h>

/** describes one input buffer for the osx writer */
struct osx_buffer {
	/** pointer to the beginning of the buffer */
	float *buffer;
	/** the size of this buffer */
	long size;
	/** current position in the buffer */
	float *ptr;
	/** number of floats not yet consuned */
	long remaining;
	/** pointer to the next audio buffer */
	struct osx_buffer *next;
};

/** data specific to the osx writer */
struct private_osx_write_data {
	/** the main control structure for audio data manipulation */
	AudioUnit audio_unit;
	/** non-zero if playback has started */
	char play;
	/** callback reads audio data from this buffer */
	struct osx_buffer *from;
	/** the post_select writes audio data here */
	struct osx_buffer *to;
	/** sample rate of the current audio stream */
	unsigned sample_rate;
	/** Sample format of the current audio stream */
	unsigned sample_format;
	/** number of channels of the current audio stream */
	unsigned channels;
};

static void destroy_buffers(struct private_osx_write_data *powd)
{
	struct osx_buffer *ptr;
	struct osx_buffer *ptr2;
	ptr = powd->to->next;
	powd->to->next = NULL;
	while (ptr) {
		ptr2 = ptr->next;
		free(ptr->buffer);
		free(ptr);
		ptr = ptr2;
	}
}

static void init_buffers(struct writer_node *wn)
{
	struct private_osx_write_data *powd = wn->private_data;
	struct osx_write_args_info *conf = wn->conf;
	struct osx_buffer **ptrptr;
	int i;

	ptrptr = &powd->to;
	for (i = 0; i < conf->numbuffers_arg; i++) {
		*ptrptr = para_malloc(sizeof(struct osx_buffer));
		(*ptrptr)->size = 0;
		(*ptrptr)->remaining = 0;
		(*ptrptr)->buffer = NULL;
		ptrptr = &(*ptrptr)->next;
	}
	*ptrptr = powd->from = powd->to;
}

static void fill_buffer(struct private_osx_write_data *powd, char *data, long bytes)
{
	struct osx_buffer *b = powd->to;
	float *dest;
	long samples;
	enum sample_format sf = powd->sample_format;

	samples = (sf == SF_S8 || sf == SF_U8)? bytes : bytes / 2;
	assert(b->remaining == 0 || samples > 0);
	if (b->size != samples) {
		b->buffer = para_realloc(b->buffer, samples * sizeof(float));
		b->size = samples;
	}
	dest = b->buffer;
	switch (powd->sample_format) {
	case SF_U8: {
		uint8_t *src = (uint8_t *)data;
		while (samples--) {
			*dest++ = (*src++) / 256.0;
		}
		break;
	}
	case SF_S8: {
		int8_t *src = (int8_t *)data;
		while (samples--) {
			*dest++ = ((*src++) + 128) / 256.0;
		}
		break;
	}
	default: {
		short *src = (short *)data;
		while (samples--)
			*dest++ = (*src++) / 32768.0;
	}
	}
	b->ptr = b->buffer;
	b->remaining = b->size;
}

static OSStatus osx_callback(void * inClientData,
	__a_unused AudioUnitRenderActionFlags *inActionFlags,
	__a_unused const AudioTimeStamp *inTimeStamp,
	__a_unused  UInt32 inBusNumber,
	__a_unused UInt32 inNumFrames,
	AudioBufferList *outOutputData)
{
	long m, n;
	float *dest;
	int i;
	struct private_osx_write_data *powd = inClientData;

//	PARA_INFO_LOG("%p\n", powd);
	for (i = 0; i < outOutputData->mNumberBuffers; ++i) {
		/* what we have to fill */
		m = outOutputData->mBuffers[i].mDataByteSize / sizeof(float);
		dest = outOutputData->mBuffers[i].mData;
		while (m > 0) {
			n = powd->from->remaining;
			if (n <= 0) {
				n = powd->from->next->remaining;
				if (n <= 0) {
					PARA_INFO_LOG("buffer underrun\n");
					return 0;
				}
				powd->from = powd->from->next;
			}
//			PARA_INFO_LOG("buf %p: n = %ld, m= %ld\n", powd->from->buffer, n, m);
			/*
			 * we dump what we can. In fact, just the necessary
			 * should be sufficient
			 */
			if (n > m)
				n = m;
			memcpy(dest, powd->from->ptr, n * sizeof(float));
			dest += n;
			/* remember all done work */
			m -= n;
			powd->from->ptr += n;
			if ((powd->from->remaining -= n) <= 0)
				powd->from = powd->from->next;
		}
	}
	return 0;
}

#ifdef WORDS_BIGENDIAN /* ppc */
#define ENDIAN_FLAGS kLinearPCMFormatFlagIsBigEndian
#else
#define ENDIAN_FLAGS 0
#endif

static int core_audio_init(struct writer_node *wn)
{
	struct private_osx_write_data *powd = para_calloc(sizeof(*powd));
	ComponentDescription desc;
	Component comp;
	AURenderCallbackStruct inputCallback = {osx_callback, powd};
	AudioStreamBasicDescription format;
	int ret;
	struct btr_node *btrn = wn->btrn;
	int32_t val;

	wn->private_data = powd;
	init_buffers(wn);
	/* where did that default audio output go? */
	desc.componentType = kAudioUnitType_Output;
	desc.componentSubType = kAudioUnitSubType_DefaultOutput;
	/* NOTE: and if default output isn't Apple? */
	desc.componentManufacturer = kAudioUnitManufacturer_Apple;
	desc.componentFlags = 0;
	desc.componentFlagsMask = 0;
	ret = -E_DEFAULT_COMP;
	comp = FindNextComponent(NULL, &desc);
	if (!comp)
		goto e0;
	ret = -E_OPEN_COMP;
	if (OpenAComponent(comp, &powd->audio_unit))
		goto e0;
	ret = -E_UNIT_INIT;
	if (AudioUnitInitialize(powd->audio_unit))
		goto e1;
	powd->play = 0;
	get_btr_sample_rate(btrn, &val);
	powd->sample_rate = val;
	get_btr_channels(btrn, &val);
	powd->channels = val;
	get_btr_sample_format(btrn, &val);
	powd->sample_format = val;
	/*
	 * Choose PCM format. We tell the Output Unit what format we're going
	 * to supply data to it. This is necessary if you're providing data
	 * through an input callback AND you want the DefaultOutputUnit to do
	 * any format conversions necessary from your format to the device's
	 * format.
	 */
	format.mFormatID = kAudioFormatLinearPCM;
	format.mFramesPerPacket = 1;
	format.mSampleRate = powd->sample_rate;
	/* flags specific to each format */
	format.mFormatFlags = kLinearPCMFormatFlagIsFloat
		| kLinearPCMFormatFlagIsPacked
		| ENDIAN_FLAGS;
	switch (powd->sample_format) {
	case SF_S8:
	case SF_U8:
		wn->min_iqs = powd->channels;
		break;
	default:
		wn->min_iqs = powd->channels * 2;
	}
	format.mBitsPerChannel = 8 * sizeof(float);
	format.mBytesPerPacket = powd->channels * sizeof(float);
	format.mBytesPerFrame = format.mBytesPerPacket;
	format.mChannelsPerFrame = powd->channels;

	ret = -E_STREAM_FORMAT;
	if (AudioUnitSetProperty(powd->audio_unit, kAudioUnitProperty_StreamFormat,
			kAudioUnitScope_Input, 0, &format,
			sizeof(AudioStreamBasicDescription)))
		goto e2;
	ret = -E_ADD_CALLBACK;
	if (AudioUnitSetProperty(powd->audio_unit, kAudioUnitProperty_SetRenderCallback,
			kAudioUnitScope_Input, 0, &inputCallback,
			sizeof(inputCallback)) < 0)
		goto e3;
	return 1;
e3:
	destroy_buffers(powd);
e2:
	AudioUnitUninitialize(powd->audio_unit);
e1:
	CloseComponent(powd->audio_unit);
e0:
	free(powd);
	wn->private_data = NULL;
	return ret;
}

__malloc static void *osx_write_parse_config_or_die(const char *options)
{
	struct osx_write_args_info *conf = para_calloc(sizeof(*conf));

	/* exits on errors */
	osx_cmdline_parser_string(options, conf, "osx_write");
	return conf;
}

static void osx_free_config(void *conf)
{
	osx_cmdline_parser_free(conf);
}

static void osx_write_close(struct writer_node *wn)
{
	struct private_osx_write_data *powd = wn->private_data;

	if (!powd)
		return;
	PARA_INFO_LOG("closing writer node %p\n", wn);
	AudioOutputUnitStop(powd->audio_unit);
	AudioUnitUninitialize(powd->audio_unit);
	CloseComponent(powd->audio_unit);
	destroy_buffers(powd);
	free(powd);
}

static void osx_write_post_select(__a_unused struct sched *s, struct task *t)
{
	struct writer_node *wn = container_of(t, struct writer_node, task);
	struct private_osx_write_data *powd = wn->private_data;
	struct btr_node *btrn = wn->btrn;
	char *data;
	size_t bytes;
	int ret = 0;

	while (!powd || powd->to->remaining <= 0) {
		ret = btr_node_status(wn->btrn, wn->min_iqs, BTR_NT_LEAF);
		if (ret <= 0)
			break;
		if (!powd) {
			ret = core_audio_init(wn);
			if (ret < 0)
				break;
			powd = wn->private_data;
		}
		btr_merge(btrn, 8192);
		bytes = btr_next_buffer(btrn, &data);
		//PARA_CRIT_LOG("have: %zu\n", bytes);
		fill_buffer(powd, data, bytes);
		btr_consume(btrn, bytes);
		if (!powd->play) {
			ret = -E_UNIT_START;
			if (AudioOutputUnitStart(powd->audio_unit))
				break;
			powd->play = 1;
		}
		powd->to = powd->to->next;
	}
	if (ret < 0 && (!powd || powd->from->remaining <= 0)) {
		btr_remove_node(btrn);
		t->error = ret;
	}
}

static void osx_write_pre_select(struct sched *s, struct task *t)
{
	struct writer_node *wn = container_of(t, struct writer_node, task);
	struct private_osx_write_data *powd = wn->private_data;
	struct timeval tmp = {.tv_sec = 1, .tv_usec = 0}, delay = tmp;
	unsigned long factor;
	size_t numbytes;
	int ret = btr_node_status(wn->btrn, wn->min_iqs, BTR_NT_LEAF);

	if (ret == 0)
		return;
	if (ret < 0 || !powd)
		return sched_min_delay(s);
	assert(powd->sample_rate > 0);
	assert(wn->min_iqs > 0);
	numbytes = powd->to->remaining * sizeof(short);
	factor = numbytes / powd->sample_rate / wn->min_iqs;
	tv_scale(factor, &tmp, &delay);
	sched_request_timeout(&delay, s);
}

/** the init function of the osx writer */
void osx_write_init(struct writer *w)
{
	struct osx_write_args_info dummy;

	osx_cmdline_parser_init(&dummy);
	w->close = osx_write_close;
	w->pre_select = osx_write_pre_select;
	w->post_select = osx_write_post_select;
	w->parse_config_or_die = osx_write_parse_config_or_die;
	w->free_config = osx_free_config;
	w->shutdown = NULL; /* nothing to do */
	w->help = (struct ggo_help) {
		.short_help = osx_write_args_info_help,
		.detailed_help = osx_write_args_info_detailed_help
	};
	osx_cmdline_parser_free(&dummy);
}
